'''Named fields'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat.model.binary import StrSerializable, BytesSerializable
from warcat.model.common import NEWLINE
import collections
import logging
import re


_logger = logging.getLogger(__name__)


class Fields(StrSerializable, BytesSerializable):
    '''Name and value pseudo-map list

    Behaves like a `dict` or mutable mapping. Mutable mapping operations
    remove any duplicates in the field list.
    '''

    def __init__(self, field_list=None):
        self._list = [] if field_list is None else field_list

    def __contains__(self, name):
        return self.get(name) is not None

    def __iter__(self):
        return self._list

    def __len__(self):
        return len(self._list)

    def __getitem__(self, name):
        for k, v in self._list:
            if k.lower() == name.lower():
                return v

        raise KeyError('{} not in fields'.format(name))

    def __setitem__(self, name, value):
        try:
            index = self.index(name)
        except KeyError:
            self._list.append((name, value))
        else:
            del self[name]
            self._list.insert(index, (name, value))

    def __delitem__(self, name):
        self._list[:] = [x for x in self._list if x[0].lower() != name.lower()]

    def add(self, name, value):
        '''Append a name-value field to the list'''
        self._list.append((name, value))

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def get_list(self, name):
        '''Return a list of values'''

        return list([x for x in self._list if x[0].lower() == name.lower()])

    def count(self, name):
        '''Count the number of times this name occurs in the list'''

        return len(self.get_list(name))

    def index(self, name):
        '''Return the index of the first occurance of given name'''

        for i in range(len(self._list)):
            if self._list[i][0].lower() == name.lower():
                return i

        raise KeyError('Name {} not found in fields'.format(name))

    def list(self):
        '''Return the underlying list'''

        return self._list

    def keys(self):
        return [x[0] for x in self._list]

    def values(self):
        return [x[1] for x in self._list]

    def clear(self):
        self._list[:] = []

    def iter_str(self):
        for name, value in self._list:
            if value:
                yield '{}: {}'.format(name, value)
            else:
                yield '{}:'.format(name)
            yield NEWLINE

    def iter_bytes(self):
        for s in self.iter_str():
            yield s.encode()

    @classmethod
    def parse(cls, s, newline=NEWLINE):
        '''Parse a named field string and return a :class:`Fields`'''

        fields = Fields()
        lines = collections.deque(
            s.split(newline) if isinstance(newline, str) else
            re.split(newline, s)
        )

        while lines:
            line = lines.popleft()

            if not line:
                continue

            name, value = line.split(':', 1)
            value = value.lstrip()
            value = cls.join_multilines(value, lines)
            fields.add(name, value)

        return fields

    @classmethod
    def join_multilines(cls, value, lines):
        '''Scan for multiline value which is prefixed with a space or tab'''

        while lines:
            line = lines.popleft()

            if not line:
                break
            if line[0] not in (' ', '\t'):
                lines.appendleft(line)
                break

            value = '{}{}'.format(value, line[1:])

        return value


class Header(StrSerializable, BytesSerializable):
    '''A header of a WARC Record.

    .. attribute:: version

        A `str` containing the version

    .. attribute:: fields

        The :class:`Fields` object.
    '''

    VERSION = '1.0'

    def __init__(self, version=VERSION, fields=None):
        self.version = version
        self.fields = fields or Fields()

    @classmethod
    def parse(cls, b):
        '''Parse from `bytes` and return :class:`Header`'''

        version_line, field_str = b.decode().split(NEWLINE, 1)

        _logger.debug('Version line=%s', version_line)

        if not version_line.startswith('WARC'):
            raise IOError('Wrong WARC header')

        header = Header(version_line[5:], Fields.parse(field_str))

        return header

    def iter_str(self):
        yield 'WARC/'
        yield self.version
        yield NEWLINE

        for v in self.fields.iter_str():
            yield v

        yield NEWLINE

    def iter_bytes(self):
        for s in self.iter_str():
            yield s.encode()


class HTTPHeader(Fields):
    '''Fields extended with a HTTP status attribute.

    .. attribute:: status

        The `str` of the HTTP status message and code.
    '''

    def __init__(self, field_list=None, status=None):
        Fields.__init__(self, field_list=field_list)
        self.status = status

    @property
    def status_code(self):
        return int(self.status.split()[1])

    @classmethod
    def parse(cls, s, newline=NEWLINE):
        http_headers = HTTPHeader()
        http_headers.status, s = s.split(newline, 1)

        fields = super(HTTPHeader, cls).parse(s, newline=newline)
        http_headers.list().extend(fields.list())

        return http_headers

    def iter_str(self):
        yield self.status
        yield NEWLINE

        for s in Fields.iter_str(self):
            yield s


HTTPHeaders = HTTPHeader
'''.. deprecated:: 2.1.1

    Name uses wrong inflection. Use :class:`HTTPHeader` instead.
'''

__all__ = ['Fields', 'HTTPHeader', 'HTTPHeaders', 'Header']



# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat.model import WARC
from warcat.tool import ListTool, ConcatTool, SplitTool, ExtractTool, VerifyTool
import argparse
import logging
import os
import sys
import warcat.version

_logger = logging.getLogger(__name__)


def main():
    arg_parser = argparse.ArgumentParser(
        description='Tool for handling Web ARChive (WARC) files.')
    arg_parser.add_argument('--version', action='version',
        version=warcat.version.__version__)
    arg_parser.add_argument('command',
        help='A command to run. Use "help" for a list.')
    arg_parser.add_argument('file', help='Filename of file to be read.',
        nargs='*')
    arg_parser.add_argument('--output', '-o', metavar='FILE',
        help='Output to FILE instead of standard out',
        type=argparse.FileType('wb'), default=sys.stdout,
    )
    arg_parser.add_argument('--gzip', '-z', action='store_true',
        help='When outputting a file, use gzip compression',
    )
    arg_parser.add_argument('--force-read-gzip', action='store_true',
        help='Instead of guessing by filename, force reading archives as'
        ' gzip compressed')
    arg_parser.add_argument('--verbose', action='count',
        help='Increase verbosity. Can be used more than once.')
    arg_parser.add_argument('--record', action='append',
        help='Apply command to record with given ID when reading. '
        'Can be used more than once.')
    arg_parser.add_argument('--preserve-block', action='store_true',
        help="Don't attempt to parse content blocks. Parsed content blocks"
        " may not match content-length and hash digests on serialization.")
    arg_parser.add_argument('--output-dir', '-d',
        default=os.getcwd(),
        help='For output operations that make multiple files, use given'
            ' directory instead of current working directory.')
    arg_parser.add_argument('--progress', action='store_true',
        help='Show progress or activity')
    arg_parser.add_argument('--keep-going', action='store_true',
        help='Continue processing records despite errors')

    original_print_help = arg_parser.print_help

    def help_monkeypatch(file=None):
        original_print_help(file)
        print(file=file)
        help_command(file=file)

    arg_parser.print_help = help_monkeypatch

    args = arg_parser.parse_args()

    if args.verbose:
        if args.verbose > 1:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)

    command_info = commands.get(args.command)

    if command_info:
        command_info[1](args)
    else:
        help_command(args)


def help_command(args=None, file=sys.stderr):
    print('Commands:', file=file)

    for command in sorted(commands):
        label = commands[command][0]
        print('{}\n    {}'.format(command, label), file=file)


def get_file_buffer(file_obj):
    if file_obj == sys.stdout:
        return sys.stdout.buffer
    else:
        return file_obj


def build_tool(class_, args):
    return class_(args.file,
        write_gzip=args.gzip,
        force_read_gzip=args.force_read_gzip,
        out_file=get_file_buffer(args.output),
        read_record_ids=args.record,
        preserve_block=args.preserve_block,
        out_dir=args.output_dir,
        print_progress=args.progress,
        keep_going=args.keep_going,
    )


def list_command(args):
    tool = build_tool(ListTool, args)
    tool.process()


def pass_command(args):
    out_file = get_file_buffer(args.output)

    for filename in args.file:
        warc = WARC()
        warc.load(filename, force_gzip=args.force_read_gzip)

        for v in warc.iter_bytes():
            out_file.write(v)


def concat_command(args):
    tool = build_tool(ConcatTool, args)
    tool.process()


def split_command(args):
    tool = build_tool(SplitTool, args)
    tool.process()


def extract_command(args):
    tool = build_tool(ExtractTool, args)
    tool.process()


def verify_command(args):
    tool = build_tool(VerifyTool, args)
    tool.process()

    if tool.problems:
        sys.exit('Validation failed. Problems: {}.'.format(tool.problems))


commands = {
    'help': ('List commands available', help_command),
    'list': ('List contents of archive', list_command),
    'pass': ('Load archive and write it back out', pass_command),
    'concat': ('Naively join archives into one', concat_command),
    'split': ('Split archives into individual records', split_command),
    'extract': ('Extract files from archive', extract_command),
    'verify': ('Verify digest and validate conformance', verify_command),
}


if __name__ == '__main__':
    main()



from warcat import util
import datetime
import io
import os.path
import unittest


class TestUtil(unittest.TestCase):
    def test_find_file_pattern(self):
        f = io.BytesIO(b'abcdefg\r\n\r\nhijklmnop')
        offset = util.find_file_pattern(f, b'\r\n\r\n', inclusive=True)
        self.assertEqual(11, offset)
        self.assertEqual(0, f.tell())

    def test_disk_buffered_reader(self):
        test_data = b'0123456789' * 100

        f = util.DiskBufferedReader(io.BytesIO(test_data), disk_buffer_size=42)

        self.assertEqual(b'0', f.peek(1))
        self.assertEqual(b'0', f.read(1))
        self.assertEqual(b'1', f.read(1))
        self.assertEqual(b'2', f.peek(1))

        f.seek(45)
        self.assertEqual(b'5', f.peek(1))
        self.assertEqual(b'56', f.read(2))

        f.seek(41)

        self.assertEqual(b'1234', f.read(4))

        f.seek(0)

        self.assertEqual(b'0', f.peek(1))
        self.assertEqual(b'0', f.read(1))
        self.assertEqual(b'1', f.read(1))
        self.assertEqual(b'2', f.peek(1))

    def test_find_file_pattern_loop_boundary(self):
        for i in range(1000):
            data = b'x' * i + b'\r\n\r\nabcdefghijklmnop'

            f = io.BytesIO(data)

            self.assertEqual(i, util.find_file_pattern(f, b'\r\n\r\n'))

    def test_split_url_to_filename(self):
        self.assertEqual(['example.com', 'index.php_article=Main_Page'],
            util.split_url_to_filename(
                'http://example.com/index.php?article=Main_Page')
        )

        def f1():
            util.split_url_to_filename('http://example.com/../system')
        self.assertRaises(ValueError, f1)

        def f2():
            util.split_url_to_filename('http://example.com/./system')
        self.assertRaises(ValueError, f2)

    def test_append_index_filename(self):
        self.assertEqual('_index_da39a3', util.append_index_filename(''))
        self.assertEqual('index.php_index_bb6499',
            util.append_index_filename('index.php'))
        self.assertEqual(os.path.join('hello', 'index.php_index_bb6499'),
            util.append_index_filename(os.path.join('hello', 'index.php'))
        )

    def test_truncate_filename_parts(self):
        self.assertEqual(
            ['a', 'b', 'c' * 160 + '_4658bd', 'd'],
            util.truncate_filename_parts(['a', 'b', 'c' * 400, 'd'])
        )

    def test_parse_http_date(self):
        self.assertEqual(datetime.datetime(1995, 11, 20, 19, 12, 8,
            tzinfo=datetime.timezone(datetime.timedelta(-1, 68400))),
            util.parse_http_date('Mon, 20 Nov 1995 19:12:08 -0500'))



'''Utility functions'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
import collections
import datetime
import email.utils
import hashlib
import http.client
import io
import logging
import os
import tempfile
import threading
import urllib.parse


_logger = logging.getLogger(__name__)


def printable_str_to_str(s):
    return s.translate(str.maketrans('', '', '\t\r\n'))\
        .replace(r'\r', '\r')\
        .replace(r'\n', '\n')\
        .replace(r'\t', '\t')


def find_file_pattern(file_obj, pattern, bufsize=512, limit=4096,
inclusive=False):
    '''Find the offset from current position of pattern'''

    original_position = file_obj.tell()
    bytes_read = 0
    # FIXME: don't accumulate growing buffer
    search_buf = io.BytesIO()

    while True:
        if limit:
            size = min(bufsize, limit - bytes_read)
        else:
            size = bufsize

        data = file_obj.read(size)

        if not data:
            break

        search_buf.write(data)

        try:
            index = search_buf.getvalue().index(pattern)
        except ValueError:
            pass
        else:
            offset = index

            if inclusive:
                offset += len(pattern)

            file_obj.seek(original_position)
            return offset

        bytes_read += len(data)

    file_obj.seek(original_position)

    raise ValueError('Search for pattern exhausted')


def strip_warc_extension(s):
    '''Removes ``.warc`` or ``.warc.gz`` from filename'''

    if s.endswith('.gz'):
        s = s[:-3]

    if s.endswith('.warc'):
        s = s[:-5]

    return s


class DiskBufferedReader(io.BufferedIOBase):
    '''Buffers the file to disk large parts at a time'''

    # Some segments lifted from _pyio.py
    # Copyright 2001-2011 Python Software Foundation
    # Licensed under Python Software Foundation License Version 2

    def __init__(self, raw, disk_buffer_size=104857600, spool_size=10485760):
        io.BufferedIOBase.__init__(self)
        self._raw = raw
        self._disk_buffer_size = disk_buffer_size
        self._offset = 0
        self._block_index = None
        self._block_file = None
        self._spool_size = spool_size
        self._lock = threading.RLock()
        self._cache = FileCache()

        self._set_block(0)

    def _set_block(self, index):
        if index == self._block_index:
            return

        with self._lock:
            self._block_index = index

            self._block_file = self._cache.get(self._block_index)

            if self._block_file:
                _logger.debug('Buffer block file cache hit. index=%d',
                    self._block_index)
            else:
                _logger.debug('Creating buffer block file. index=%d',
                    self._block_index)

                self._block_file = tempfile.SpooledTemporaryFile(
                    max_size=self._spool_size)

                self._raw.seek(self._block_index * self._disk_buffer_size)
                copyfile_obj(self.raw, self._block_file,
                    max_length=self._disk_buffer_size)
                self._cache.put(self._block_index, self._block_file)

                _logger.debug('Buffer block file created. length=%d',
                    self._block_file.tell())

            self._block_file.seek(0)

    def tell(self):
        return self._offset

    def seek(self, pos, whence=0):
        if not (0 <= whence <= 1):
            raise ValueError('Bad whence argument')

        with self._lock:
            if whence == 1:
                self._offset += pos

            self._offset = pos
            index = self._offset // self._disk_buffer_size
            self._set_block(index)
            self._block_file.seek(self._offset % self._disk_buffer_size)

    def read(self, n=None):
        buf = io.BytesIO()
        bytes_left = n

        with self._lock:
            while True:
                self.seek(self._offset)
                data = self._block_file.read(bytes_left)
                self._offset += len(data)
                buf.write(data)
                bytes_left -= len(data)

                if not data:
                    break

                if bytes_left <= 0:
                    break

        return buf.getvalue()

    def peek(self, n=0):
        with self._lock:
            original_position = self.tell()
            data = self.read(n)
            self.seek(original_position)
        return data

    def seekable(self):
        return self.raw.seekable()

    def readable(self):
        return self.raw.readable()

    def writable(self):
        return False

    @property
    def raw(self):
        return self._raw

    @property
    def closed(self):
        return self.raw.closed

    @property
    def name(self):
        return self.raw.name

    @property
    def mode(self):
        return self.raw.mode

    def fileno(self):
        return self.raw.fileno()

    def isatty(self):
        return self.raw.isatty()


class FileCache(object):
    '''A cache containing references to file objects.

    File objects are closed when expired. Class is thread safe and will
    only return file objects belonging to its own thread.
    '''

    def __init__(self, size=4):
        self._size = size
        self._files = collections.deque()
        self._lock = threading.Lock()

    def get(self, filename):
        thread_id = threading.current_thread()

        with self._lock:
            return self._get(filename, thread_id)

    def _get(self, filename, thread_id):
        for cache_filename, cache_thread_id, file_obj in self._files:
            if filename == cache_filename and thread_id == cache_thread_id:
                return file_obj

    def put(self, filename, file_obj):
        thread_id = threading.current_thread()

        with self._lock:
            if self._get(filename, thread_id):
                return

            if len(self._files) > self._size:
                old_file_obj = self._files.popleft()[2]
                old_file_obj.close()

            self._files.append((filename, thread_id, file_obj))


def copyfile_obj(source, dest, bufsize=4096, max_length=None,
write_attr_name='write'):
    '''Like :func:`shutil.copyfileobj` but with limit on how much to copy'''

    bytes_read = 0
    write_func = getattr(dest, write_attr_name)

    while True:
        if max_length != None:
            read_size = min(bufsize, max_length - bytes_read)
        else:
            read_size = bufsize

        data = source.read(read_size)

        if not data:
            break

        write_func(data)
        bytes_read += len(data)


class HTTPSocketShim(io.BytesIO):
    def makefile(self, *args, **kwargs):
        return self


def parse_http_response(file_obj):
    '''Parse and return :class:`http.client.HTTPResponse`'''

    response = http.client.HTTPResponse(HTTPSocketShim(file_obj))
    response.begin()

    return response


def split_url_to_filename(s):
    '''Attempt to split a URL to a filename on disk'''

    url_info = urllib.parse.urlsplit(s)

    l = [sanitize_str(url_info.netloc)]

    for part in url_info.path.lstrip('/').split('/'):
        part = sanitize_str(part)

        l.append(part)

        if not part:
            l[-1] = append_index_filename(part)

    if url_info.query:
        l[-1] += '_' + sanitize_str(url_info.query)

    if frozenset([os.curdir, os.pardir, '.', '..']) & frozenset(l):
        raise ValueError('Path contains directory traversal filenames')

    return l


SANITIZE_BLACKLIST = frozenset(
    r'/\:*?"<>|' + ''.join([chr(i) for i in range(0, 32)]) + '\x7f'
)


def sanitize_str(s):
    '''Replaces unsavory chracters from string with an underscore'''

    return ''.join([c if c not in SANITIZE_BLACKLIST else '_' for c in s])


def append_index_filename(path):
    '''Adds ``_index_xxxxxx`` to the path.

    It uses the basename aka filename of the path to generate the hex hash
    digest suffix.
    '''

    hasher = hashlib.sha1(os.path.basename(path).encode())
    path += '_index_{}'.format(hasher.hexdigest()[:6])

    return path


def rename_filename_dirs(dest_filename):
    '''Renames files if they conflict with a directory in given path.

    If a file has the same name as the directory, the file is renamed
    using :func:`append_index_filename`.
    '''

    path = dest_filename
    while True:
        path, filename = os.path.split(path)

        if not filename:
            break

        if os.path.isfile(path):
            new_path = append_index_filename(path)

            _logger.debug('Rename %s -> %s', path, new_path)
            os.rename(path, new_path)
            break


def truncate_filename_parts(path_parts, length=160):
    '''Truncate and suffix filename path parts if they exceed the given length.

    If the filename part is too long, the part is truncated and an underscore
    plus a 6 letter hex (_xxxxxx) suffix is appended.
    '''
    new_parts = list(path_parts)

    for index in range(len(path_parts)):
        part = path_parts[index]

        if len(part) > length:
            hasher = hashlib.sha1(part.encode())
            new_part = '{}_{}'.format(part[:length], hasher.hexdigest()[:6])
            new_parts[index] = new_part

    return new_parts


def parse_http_date(s):
    t = email.utils.parsedate_tz(s)

    if not t:
        raise ValueError('Unable to parse date')

    tzinfo = datetime.timezone(datetime.timedelta(seconds=t[9]))
    d = datetime.datetime(*t[:6], tzinfo=tzinfo)

    return d


file_cache = FileCache()
'''The :class:`FileCache` instance'''



'''Content blocks and payload blocks'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat import util
from warcat.model.binary import BytesSerializable, BinaryFileRef
from warcat.model.common import FIELD_DELIM_BYTES, NEWLINE_BYTES
from warcat.model.field import HTTPHeader, Fields
import logging


_logger = logging.getLogger(__name__)


class ContentBlock(BytesSerializable):
    @classmethod
    def load(cls, file_obj, length, content_type):
        '''Load and return :class:`BinaryBlock` or :class:`BlockWithPayload`'''

        if content_type.startswith('application/http'):
            return BlockWithPayload.load(file_obj, length,
                field_cls=HTTPHeader, strict=False)
        elif content_type.startswith('application/warc-fields'):
            return BlockWithPayload.load(file_obj, length, field_cls=Fields)
        else:
            return BinaryBlock.load(file_obj, length)


class BinaryBlock(ContentBlock, BinaryFileRef):
    '''A content block that is octet data'''

    def iter_bytes(self):
        for v in self.iter_file():
            yield v

    @classmethod
    def load(cls, file_obj, length):
        '''Return a :class:`BinaryBlock` using given file object'''

        binary_block = BinaryBlock()
        binary_block.set_file(file_obj.name or file_obj, offset=file_obj.tell(),
            length=length)

        file_obj.seek(file_obj.tell() + length)

        _logger.debug('Binary content block length=%d', binary_block.length)

        return binary_block


class BlockWithPayload(ContentBlock):
    '''A content block (fields/data) within a :class:`Record`.

    .. attribute:: fields

        :class:`Fields`

    .. attribute:: payload

        :class:`Payload`

    .. attribute:: binary_block

        If this block was loaded from a file, this attribute will be a
        :class:`BinaryBlock` of the original file. Otherwise, this attribute
        is `None`.
    '''

    def __init__(self, fields=None, payload=None):
        self.fields = fields or Fields()
        self.payload = payload or Payload()
        self.binary_block = None

    @classmethod
    def load(cls, file_obj, length, field_cls, strict=True):
        '''Return a :class:`BlockWithPayload`

        :param file_obj: The file object
        :param length: How much to read from the file
        :param field_cls: The class or subclass of :class:`Fields`
        '''

        binary_block = BinaryBlock()
        binary_block.set_file(file_obj.name or file_obj, file_obj.tell(), length)

        try:
            field_length = util.find_file_pattern(file_obj, FIELD_DELIM_BYTES,
                limit=length, inclusive=True)
        except ValueError:
            # No payload
            field_length = length

        errors = 'strict' if strict else 'replace'
        field_str = file_obj.read(field_length).decode(errors=errors)
        fields = field_cls.parse(field_str)
        payload_length = length - field_length
        payload = Payload()

        payload.set_file(file_obj.name or file_obj, offset=file_obj.tell(),
            length=payload_length)
        _logger.debug('Field length=%d', field_length)
        _logger.debug('Payload length=%d', payload_length)

        file_obj.seek(file_obj.tell() + payload_length)

        block = BlockWithPayload(fields, payload)
        block.binary_block = binary_block

        return block

    @property
    def length(self):
        '''Return the new computed length'''

        return (len(bytes(self.fields)) + len(NEWLINE_BYTES) +
            self.payload.length)

    def iter_bytes(self):
        for v in self.fields.iter_bytes():
            yield v

        yield NEWLINE_BYTES

        for v in self.payload.iter_bytes():
            yield v


class Payload(BytesSerializable, BinaryFileRef):
    '''Data within a content block that has fields'''

    def __init__(self):
        BinaryFileRef.__init__(self)

    def iter_bytes(self):
        for v in self.iter_file():
            yield v


__all__ = ['ContentBlock', 'BinaryBlock', 'BlockWithPayload', 'Payload']



'''A WARC record'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat import util
from warcat.model.binary import BytesSerializable
from warcat.model.block import ContentBlock, BinaryBlock
from warcat.model.common import FIELD_DELIM_BYTES, NEWLINE_BYTES
from warcat.model.field import Header
import isodate
import logging


_logger = logging.getLogger(__name__)


class Record(BytesSerializable):
    '''A WARC Record within a WARC file.


    .. attribute:: header

        :class:`Header`

    .. attribute:: content_block

        A :class:`BinaryBlock` or :class:`BlockWithPayload`

    .. attribute:: file_offset

        If this record was loaded from a file, this attribute contains
        an `int` describing the location of the record in the file.
    '''

    def __init__(self, header=None, content_block=None):
        self.header = header or Header()
        self.content_block = None
        self.file_offset = None

    @classmethod
    def load(cls, file_obj, preserve_block=False, check_block_length=True):
        '''Parse and return a :class:`Record`

        :param file_object: A file-like object.
        :param preserve_block: If `True`, content blocks are not parsed
            for fields and payloads. Enabling this feature ensures
            preservation of content length and hash digests.
        :param check_block_length: If `True`, the length of the blocks are
            checked to a serialized version by Warcat. This can be useful for
            checking whether Warcat will output blocks with correct
            whitespace.
        '''

        _logger.debug('Record start at %d 0x%x', file_obj.tell(),
            file_obj.tell())

        record = Record()
        record.file_offset = file_obj.tell()
        header_length = util.find_file_pattern(file_obj, FIELD_DELIM_BYTES,
            inclusive=True)
        record.header = Header.parse(file_obj.read(header_length))
        block_length = record.content_length

        _logger.debug('Block length=%d', block_length)

        if not preserve_block:
            content_type = record.header.fields.get('content-type')
            record.content_block = ContentBlock.load(file_obj, block_length,
                content_type)
        else:
            record.content_block = BinaryBlock.load(file_obj, block_length)

        if check_block_length:
            new_content_length = record.content_block.length

            if block_length != new_content_length:
                _logger.warn('Content block length changed from %d to %d',
                    record.content_length, new_content_length)
                record.content_length = new_content_length

        return record

    @property
    def record_id(self):
        return self.header.fields['WARC-Record-ID']

    @record_id.setter
    def record_id(self, s):
        self.header.fields['WARC-Record-ID'] = s

    @property
    def content_length(self):
        return int(self.header.fields['Content-Length'])

    @content_length.setter
    def content_length(self, i):
        self.header.fields['Content-Length'] = int(i)

    @property
    def date(self):
        return isodate.parse_datetime(self.header.fields['WARC-Date'])

    @date.setter
    def date(self, datetime_obj):
        self.header.fields['WARC-Date'] = isodate.datetime_isoformat(
            datetime_obj)

    @property
    def warc_type(self):
        return self.header.fields['WARC-Type']

    @warc_type.setter
    def warc_type(self, s):
        self.header.fields['WARC-Type'] = s

    def iter_bytes(self):
        _logger.debug('Iter bytes on record %s', self.record_id)

        for v in self.header.iter_bytes():
            yield v

        if self.content_block:
            for v in self.content_block.iter_bytes():
                yield v

        yield NEWLINE_BYTES
        yield NEWLINE_BYTES

__all__ = ['Record']



'''Constants and things'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.


NEWLINE = '\r\n'
'''String CR LF'''

NEWLINE_BYTES = b'\r\n'
'''Bytes CR LF'''

FIELD_DELIM_BYTES = NEWLINE_BYTES * 2
'''Bytes CR LF CR LF'''



'''Model serialization and binary references'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat import util
import abc
import gzip
import logging
import tempfile


_logger = logging.getLogger(__name__)


class BytesSerializable(metaclass=abc.ABCMeta):
    '''Metaclass that indicates this object can be serialized to bytes'''

    @abc.abstractmethod
    def iter_bytes(self):
        '''Return an iterable of bytes'''
        pass

    def __bytes__(self):
        return b''.join(self.iter_bytes())


class StrSerializable(metaclass=abc.ABCMeta):
    '''Metaclass that indicates this object can be serialized to str'''

    @abc.abstractmethod
    def iter_str(self):
        '''Return an iterable of str'''
        pass

    def __str__(self):
        return ''.join(self.iter_str())


class BinaryFileRef(metaclass=abc.ABCMeta):
    '''Reference to a file containing the content block data.

    .. attribute:: file_offset

        When reading, the file is seeked to `file_offset`.

    .. attribute:: length

        The length of the data

    .. attribute:: filename

        The filename of the referenced data. It must be a valid file.

    .. attribute:: file_obj

        The file object to be read from. It is important that this file
        object is not shared or race conditions will occur. File objects
        are not closed automatically.

    .. note::

        Either :attr:`filename` or :attr:`file_obj` must be set.
    '''

    def __init__(self):
        self.file_offset = 0
        self.length = None
        self.filename = None
        self.file_obj = None

    def set_file(self, file, offset=0, length=None):
        '''Set the reference to the file or filename with the data.

        This is a convenience function to setting the attributes individually.
        '''

        assert file

        if hasattr(file, 'read'):
            self.file_obj = file
        else:
            self.filename = file

        self.file_offset = offset
        self.length = length

    def iter_file(self, buffer_size=4096):
        '''Return an iterable of bytes of the source data'''

        with self.get_file(safe=True) as file_obj:
            bytes_read = 0

            while True:
                if self.length is not None:
                    length = min(buffer_size, self.length - bytes_read)
                else:
                    length = buffer_size

                data = file_obj.read(length)
                bytes_read += len(data)

                if not data or not length:
                    break

                yield data

    def get_file(self, safe=True, spool_size=10485760):
        '''Return a file object with the data.

        :param safe:
            If `True`, return a new file object that is a copy of the data.
            You will be responsible for closing the file.

            Otherwise, it will be the original file object that is seeked
            to the correct offset. Be sure to not read beyond its length and
            seek back to the original position if necessary.
        '''

        if self.filename:
            file_obj = util.file_cache.get(self.filename)

            if not file_obj:
                if self.filename.endswith('.gz'):
                    file_obj = util.DiskBufferedReader(
                        gzip.GzipFile(self.filename))
                else:
                    file_obj = open(self.filename, 'rb')

                util.file_cache.put(self.filename, file_obj)
        else:
            file_obj = self.file_obj

        original_position = file_obj.tell()

        if self.file_offset:
            file_obj.seek(self.file_offset)

        if safe:
            _logger.debug('Creating safe file of %s',
                self.filename or self.file_obj)
            temp_file_obj = tempfile.SpooledTemporaryFile(max_size=spool_size)

            util.copyfile_obj(file_obj, temp_file_obj, max_length=self.length)
            temp_file_obj.seek(0)
            file_obj.seek(original_position)

            return temp_file_obj

        return file_obj


__all__ = ['BytesSerializable', 'StrSerializable', 'BinaryFileRef']



'''Archive process tools'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat import model, util, verify
import abc
import gzip
import http.client
import isodate
import itertools
import logging
import os.path
import shutil
import sys
import time


_logger = logging.getLogger(__name__)


THROBBER = [
    '(=   |',
    '|=   |',
    '| =  |',
    '|  = |',
    '|   =)',
    '|   =|',
    '|  = |',
    '| =  |',
]
# lol; so bouncy.


class VerifyProblem(ValueError):
    def __init__(self, message, iso_section=None, major=True):
        ValueError.__init__(self, message, iso_section, major)

    @property
    def message(self):
        return self.args[0]

    @property
    def iso_section(self):
        return self.args[1]

    @property
    def major(self):
        return self.args[2]


class BaseIterateTool(metaclass=abc.ABCMeta):
    '''Base class for iterating through records'''

    def __init__(self, filenames, out_file=None, write_gzip=False,
    force_read_gzip=None, read_record_ids=None, preserve_block=True,
    out_dir=None, print_progress=False, keep_going=False):
        if not out_file:
            try:
                out_file = sys.stdout.buffer
            except AttributeError:
                out_file = sys.stdout

        self.filenames = filenames
        self.out_file = out_file
        self.force_read_gzip = force_read_gzip
        self.write_gzip = write_gzip
        self.current_filename = None
        self.read_record_ids = read_record_ids
        self.preserve_block = preserve_block
        self.out_dir = out_dir
        self.print_progress = print_progress
        self.keep_going = keep_going
        self.check_block_length = False

    def preprocess(self):
        pass

    def postprocess(self):
        pass

    def process(self):
        self.num_records = 0
        throbber_iter = itertools.cycle(THROBBER)
        progress_msg = ''
        self.preprocess()

        for filename in self.filenames:
            self.record_order = 0
            self.current_filename = filename

            f = model.WARC.open(filename, force_gzip=self.force_read_gzip)

            while True:
                record, has_more = model.WARC.read_record(f,
                    preserve_block=self.preserve_block,
                    check_block_length=self.check_block_length)

                skip = False

                if self.read_record_ids:
                    if record.record_id not in self.read_record_ids:
                        skip = True

                if skip:
                    _logger.debug('Skipping %s due to filter',
                        record.record_id)
                else:
                    try:
                        self.action(record)
                    except Exception as e:
                        if self.keep_going:
                            _logger.exception('Error on record %s',
                                record.record_id)
                        else:
                            raise

                if self.print_progress and self.num_records % 100 == 0:
                    s = next(throbber_iter)
                    sys.stderr.write('\b' * len(progress_msg))
                    progress_msg = '{} {} '.format(self.num_records, s)
                    sys.stderr.write(progress_msg)
                    sys.stderr.flush()

                self.record_order += 1
                self.num_records += 1

                if not has_more:
                    break

            f.close()

        self.postprocess()

        if self.print_progress:
            sys.stderr.write('\nDone. {} records processed.\n'.format(
                self.num_records))

    @abc.abstractmethod
    def action(self, record):
        pass


class ListTool(BaseIterateTool):
    def action(self, record):
        print('Record:', record.record_id)
        print('  Order:', self.num_records)
        print('  File offset:', record.file_offset)
        print('  Type:', record.warc_type)
        print('  Date:', isodate.datetime_isoformat(record.date))
        print('  Size:', record.content_length)


class ConcatTool(BaseIterateTool):
    def preprocess(self):
        self.bytes_written = 0

    def action(self, record):
        if self.write_gzip:
            f = gzip.GzipFile(fileobj=self.out_file, mode='wb')
        else:
            f = self.out_file

        for v in record.iter_bytes():
            _logger.debug('Wrote %d bytes', len(v))
            f.write(v)
            self.bytes_written += len(v)

        if self.write_gzip:
            f.close()

        if self.num_records % 1000 == 0:
            _logger.info('Wrote %d records (%d bytes) so far',
                self.num_records, self.bytes_written)


class SplitTool(BaseIterateTool):
    def action(self, record):
        record_filename = '{}.{:08d}.warc'.format(
            util.strip_warc_extension(os.path.basename(self.current_filename)),
            self.record_order)
        record_filename = os.path.join(self.out_dir, record_filename)

        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir, exist_ok=True)

        if self.write_gzip:
            record_filename += '.gz'
            f = gzip.GzipFile(record_filename, mode='wb')
        else:
            f = open(record_filename, 'wb')

        for v in record.iter_bytes():
            _logger.debug('Wrote %d bytes', len(v))
            f.write(v)

        f.close()

        if self.num_records % 1000 == 0:
            _logger.info('Wrote %d records so far', self.num_records)


class ExtractTool(BaseIterateTool):
    def action(self, record):
        if record.warc_type != 'response':
            return
        if not isinstance(record.content_block, model.BlockWithPayload):
            return
        if not isinstance(record.content_block.fields, model.HTTPHeaders):
            return
        if not record.content_block.fields.status_code == http.client.OK:
            return

        url = record.header.fields['WARC-Target-URI']
        binary_block = record.content_block.binary_block
        file_obj = binary_block.get_file()
        data = file_obj.read(binary_block.length)
        response = util.parse_http_response(data)
        path_list = util.split_url_to_filename(url)
        path_list = util.truncate_filename_parts(path_list)
        path = os.path.join(self.out_dir, *path_list)
        dir_path = os.path.dirname(path)

        if os.path.isdir(path):
            path = util.append_index_filename(path)

        _logger.debug('Extracting %s to %s', record.record_id, path)
        util.rename_filename_dirs(path)
        os.makedirs(dir_path, exist_ok=True)

        try:
            with open(path, 'wb') as f:
                shutil.copyfileobj(response, f)
        except http.client.IncompleteRead as error:
            _logger.warning('Malformed HTTP response: %s', error)

            with open(path, 'wb') as f:
                f.write(error.partial)

        last_modified_str = response.getheader('Last-Modified')

        if last_modified_str:
            try:
                last_modified = util.parse_http_date(last_modified_str)
            except ValueError:
                pass
            else:
                timestamp = time.mktime(last_modified.utctimetuple())
                os.utime(path, (time.time(), timestamp))
                _logger.debug('Apply mtime %d to %s', timestamp, path)

        _logger.info('Extracted %s to %s', record.record_id, path)


class VerifyTool(BaseIterateTool):
    MANDATORY_FIELDS = ['WARC-Record-ID', 'Content-Length', 'WARC-Date',
        'WARC-Type']

    def preprocess(self):
        self.record_ids = set()
        self.problems = 0
        self.check_block_length = True

    def action(self, record):
        verify_actions = [
            self.verify_mandatory_fields,
#            self.check_transfer_encoding,
            self.verify_block_digest,
            self.verify_payload_digest,
            self.verify_id_uniqueness,
            self.verify_id_no_whitespace,
            self.verify_content_type,
            self.verify_concurrent_to,
            self.verify_refers_to,
            self.verify_target_uri,
            self.verify_filename,
            self.verify_profile,
            self.verify_segment_origin_id,
            self.verify_segment_total_length,
        ]

        for action in verify_actions:
            try:
                action(record)
            except VerifyProblem:
                self.problems += 1
                _logger.exception('Record %s failed validation',
                    record.record_id)

    def verify_block_digest(self, record):
        if 'WARC-Block-Digest' in record.header.fields:
            if not verify.verify_block_digest(record):
                raise VerifyProblem('Bad block digest.', '5.8')

            _logger.debug('Block digest ok')

    def verify_payload_digest(self, record):
        if 'WARC-Payload-Digest' in record.header.fields:
            if not verify.verify_payload_digest(record):
                raise VerifyProblem('Bad payload digest.', '5.9')

            _logger.debug('Payload digest ok')

    def verify_id_uniqueness(self, record):
        if record.record_id in self.record_ids:
            raise VerifyProblem('Duplicate record ID.')

        self.record_ids.add(record.record_id)

    def verify_id_no_whitespace(self, record):
        if ' ' in record.record_id:
            raise VerifyProblem('Whitespace in ID', '5.2')

    def check_transfer_encoding(self, record):
        if not isinstance(record.content_block, model.BlockWithPayload):
            return

        if 'Transfer-encoding' in record.content_block.fields:
            raise VerifyProblem('Transfer-encoding found', '5.3.2', False)

    def verify_mandatory_fields(self, record):
        for name in self.MANDATORY_FIELDS:
            if name not in record.header.fields:
                raise VerifyProblem(
                    'Mandatory {} field is missing'.format(name))

    def verify_content_type(self, record):
        if record.warc_type != 'continuation':
            return

        if record.content_length \
        and 'Content-Type' not in record.header.fields:
            raise VerifyProblem('Content-Type should be specified', '5.6',
                False)

    def verify_concurrent_to(self, record):
        if 'WARC-Concurrent-To' not in record.header.fields:
            return

        record_id = record.header.fields['WARC-Concurrent-To']

        if record.warc_type in ('warcinfo', 'conversion', 'continuation'):
            raise VerifyProblem('Unexpected WARC-Concurrent-To', '5.7')

        if record_id not in self.record_ids:
            raise VerifyProblem('Concurrent Record ID {} not seen yet'.format(
                record_id), major=False)

    def verify_refers_to(self, record):
        if 'WARC-Refers-To' not in record.header.fields:
            return

        record_id = record.header.fields['WARC-Refers-To']

        if record.warc_type in ('warcinfo', 'response', 'request',
        'continuation'):
            raise VerifyProblem('WARC-Refers-To field unexpected', '5.11')

        if record_id not in self.record_ids:
            raise VerifyProblem('Refer to record ID {} not seen yet'.format(
                record_id), major=False)

    def verify_target_uri(self, record):
        uri = record.header.fields.get('WARC-Target-URI')

        if not uri and record.warc_type in ('response', 'resource', 'request',
        'revisit', 'conversion', 'continuation'):
            raise VerifyProblem('Expected WARC-Target-URI', '5.12')

        if uri and record.warc_type == 'warc_info':
            raise VerifyProblem('Unexpected WARC-Target-URI', '5.12')

        if uri and ' ' in uri:
            raise VerifyProblem('Whitespace in URI', '5.12')

    def verify_warcinfo_id(self, record):
        record_id = record.header.fields.get('WARC-Warcinfo-ID')

        if record_id and record.warc_type == 'warcinfo':
            raise VerifyProblem('Unexpected WARC-Warcinfo-ID', '5.14')

        if not record_id:
            raise VerifyProblem('Expected WARC-Warcinfo-ID', '5.14', False)

    def verify_filename(self, record):
        if 'WARC-Filename' in record.header.fields \
        and record.warc_type != 'warcinfo':
            raise VerifyProblem('Unexpected WARC-Filename', '5.15')

    def verify_profile(self, record):
        if record.warc_type == 'revisit' \
        and 'WARC-Profile' not in record.header.fields:
            raise VerifyProblem('Expected WARC-Profile', '5.16')

    def verify_segment_origin_id(self, record):
        if record.warc_type == 'continuation':
            if 'WARC-Segment-Origin-ID' not in record.header.fields:
                raise VerifyProblem('Expected WARC-Segment-Origin-ID', '5.19')
        elif 'WARC-Segment-Origin-ID' in record.header.fields:
            raise VerifyProblem('Unexpected WARC-Segment-Origin-ID', '5.19')

    def verify_segment_total_length(self, record):
        if record.warc_type == 'continuation':
            if 'WARC-Segment-Total-Length' not in record.header.fields:
                raise VerifyProblem('Expected WARC-Segment-Total-Length',
                    '5.20')
        elif 'WARC-Segment-Total-Length' in record.header.fields:
            raise VerifyProblem('Unexpected WARC-Segment-Total-Length', '5.20')



'''WARC model starting point'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from warcat import util
from warcat.model.binary import BytesSerializable
from warcat.model.common import FIELD_DELIM_BYTES
from warcat.model.record import Record
import gzip
import logging


_logger = logging.getLogger(__name__)


class WARC(BytesSerializable):
    '''A Web ARChive file model.

    Typically, large streaming operations should use :func:`open` and
    :func:`read_record` functions.
    '''

    def __init__(self):
        self.records = []

    def load(self, filename):
        '''Open and load the contents of the given filename.

        The records are located in :attr:`records`.
        '''

        f = self.open(filename)
        self.read_file_object(f)
        f.close()

    def read_file_object(self, file_object):
        '''Read records until the file object is exhausted'''

        while True:
            record, has_more = self.read_record(file_object)
            self.records.append(record)
            if not has_more:
                break

    @classmethod
    def open(cls, filename, force_gzip=False):
        '''Return a logical file object.

        :param filename: The path of the file. gzip compression is detected
            using file extension.
        :param force_gzip: Use gzip compression always.
        '''

        if filename.endswith('.gz') or force_gzip:
            f = gzip.open(filename)
            _logger.info('Opened gziped file %s', filename)
            return util.DiskBufferedReader(f)
        else:
            f = open(filename, 'rb')
            _logger.info('Opened file %s', filename)
            return f

    @classmethod
    def read_record(cls, file_object, preserve_block=False,
    check_block_length=True):
        '''Return a record and whether there are more records to read.

        .. seealso:: :class:`Record`

        :return: A tuple. The first item is the :class:`Record`. The second
            item is a boolean indicating whether there are more records to
            be read.
        '''

        record = Record.load(file_object, preserve_block=preserve_block,
            check_block_length=check_block_length)
        _logger.debug('Finished reading a record %s', record.record_id)

        data = file_object.read(len(FIELD_DELIM_BYTES))

        if data != FIELD_DELIM_BYTES:
            _logger.debug('Wrong delim %s', data)
            raise IOError('Blocks not separated correctly (tell={})'.format(
                file_object.tell()))

        if not file_object.peek(1):
            _logger.info('Finished reading Warc')
            return (record, False)
        else:
            return (record, True)

    def iter_bytes(self):
        for record in self.records:
            for v in record.iter_bytes():
                yield v


__all__ = ['WARC']



from warcat.tool import ListTool, VerifyTool, SplitTool, ExtractTool, ConcatTool
import glob
import os.path
import tempfile
import unittest


class TestTool(unittest.TestCase):
    test_dir = os.path.join('example')

    def test_list(self):
        tool = ListTool([os.path.join(self.test_dir, 'at.warc')])
        tool.process()

    def test_verify(self):
        tool = VerifyTool([os.path.join(self.test_dir, 'at.warc')],
            preserve_block=False)
        tool.process()

        self.assertEqual(1, tool.problems)

    def test_split(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            tool = SplitTool([os.path.join(self.test_dir, 'at.warc')],
                out_dir=temp_dir)
            tool.process()

            self.assertEqual(8, len(os.listdir(temp_dir)))

    def test_extract(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            tool = ExtractTool([os.path.join(self.test_dir, 'at.warc')],
                out_dir=temp_dir, preserve_block=False)
            tool.process()

            self.assertEqual(1, len(
                glob.glob(os.path.join(temp_dir, '*', '*index*'))))

    def test_extract_long_url(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            tool = ExtractTool([os.path.join(self.test_dir, 'long_url.warc')],
                out_dir=temp_dir, preserve_block=False)
            tool.process()

            self.assertEqual(1, len(
                glob.glob(os.path.join(temp_dir, '*', '*index*'))))

            files = list(glob.glob(os.path.join(temp_dir, '*', '*index*')))
            filename = files[0].rsplit('/', 1)[-1]
            self.assertLess(len(filename), 180)

    def test_extract_bad_http_chunked_content(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            tool = ExtractTool([os.path.join(self.test_dir, 'bad_http_chunked_content.warc')],
                out_dir=temp_dir, preserve_block=False)
            tool.process()

            self.assertEqual(1, len(
                glob.glob(os.path.join(temp_dir, '*', '*index*'))))

    def test_extract_not_utf8_http_header(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            tool = ExtractTool([os.path.join(self.test_dir, 'not_utf8_http_header.warc')],
                               out_dir=temp_dir, preserve_block=False)
            tool.process()

    def test_concat(self):
        with tempfile.NamedTemporaryFile() as f:
            tool = ConcatTool([os.path.join(self.test_dir, 'at.warc')],
                out_file=f)
            tool.process()

            f.seek(0)

            tool = VerifyTool([f.name], preserve_block=False)
            tool.process()

            self.assertEqual(1, tool.problems)




'''Document model'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
from .binary import *
from .block import *
from .common import *
from .field import *
from .record import *
from .warc import *



'''Verification helpers'''
# Copyright 2013 Christopher Foo <chris.foo@gmail.com>
# Licensed under GPLv3. See COPYING.txt for details.
import hashlib
import base64
from warcat import util, model
import binascii


ALGORITHM_MAP = {
    'md5': hashlib.md5,
    'sha1': hashlib.sha1,
    'sha224': hashlib.sha224,
    'sha256': hashlib.sha256,
    'sha384': hashlib.sha384,
    'sha512': hashlib.sha512,
}


def parse_digest_field(s):
    '''Return the algorithm name and digest `bytes`'''

    algorithm, digest = s.split(':', 1)
    algorithm = algorithm.lower()
    enc_digest = digest.encode()

    try:
        digest_bytes = base64.b64decode(enc_digest)
        digest_bytes = base64.b32decode(enc_digest)
        digest_bytes = base64.b16decode(enc_digest)
    except binascii.Error as e:
        if not digest_bytes:
            raise e

    return algorithm, digest_bytes


def verify_block_digest(record):
    '''Return `True` if the content block hash digest is valid'''

    value = record.header.fields['WARC-Block-Digest']
    alg_name, given_digest = parse_digest_field(value)
    hash_obj = ALGORITHM_MAP[alg_name]()

    if isinstance(record.content_block, model.BlockWithPayload):
        content_block = record.content_block.binary_block
    else:
        content_block = record.content_block

    util.copyfile_obj(content_block.get_file(), hash_obj,
        max_length=content_block.length, write_attr_name='update')

    return given_digest == hash_obj.digest()


def verify_payload_digest(record):
    '''Return `True` if the payload hash digest is valid'''

    value = record.header.fields['WARC-Payload-Digest']
    alg_name, given_digest = parse_digest_field(value)
    hash_obj = ALGORITHM_MAP[alg_name]()
    content_block = record.content_block.payload

    util.copyfile_obj(content_block.get_file(), hash_obj,
        max_length=content_block.length, write_attr_name='update')

    return given_digest == hash_obj.digest()



import gzip

from warcat import model, util
import io
import os.path
import unittest


fields_str = util.printable_str_to_str(
r'''WARC-Type: response\r\n
Non-ASCII:    ðëARCHIVE#:>Þ   \r\n
Multiline: The quick brown fox\r\n
 jumps\r\n
\tover\n
   the lazy dog.\r\n
Content-LENGTH: 10\r\n
''')


class TestModel(unittest.TestCase):
    def test_fields_parse(self):
        fields = model.Fields.parse(fields_str)

        self.assertEqual('response', fields['warc-type'])
        self.assertEqual('ðëARCHIVE#:>Þ   ', fields['non-ascii'])
        self.assertEqual(
            'The quick brown foxjumpsover\n   the lazy dog.',
            fields['multiline'])
        self.assertEqual('10', fields['content-length'])

    def test_fields(self):
        fields = model.Fields()
        fields.add('My-Name', 'a')
        fields['Animal'] = 'kitten'
        fields.add('my-name', 'b')

        self.assertListEqual(
            [('My-Name', 'a'), ('Animal', 'kitten'), ('my-name', 'b')],
            fields.list())
        self.assertIn('my-name', fields)
        self.assertNotIn('content-length', fields)
        self.assertEqual('a', fields['my-name'])
        self.assertEqual(2, fields.count('my-name'))

        fields['my-name'] = 'c'

        self.assertEqual(1, fields.count('my-name'))
        self.assertEqual('kitten', fields['animal'])

    def test_build_model(self):
        warc = model.WARC()

        record = model.Record()
        record.content_type = 'text/plain'
        record.record_id = 'a'
        record.content_block = model.BinaryBlock()
        content_data = b'hello world!'
        record.content_block.set_file(io.BytesIO(content_data))
        record.content_length = len(content_data)
        warc.records.append(record)

        record = model.Record()
        record.content_type = 'application/warc-fields'
        record.record_id = 'b'
        record.content_block = model.BlockWithPayload()
        record.content_block.fields['example'] = 'kitten'
        content_data = b'hello world!'
        record.content_block.payload.set_file(io.BytesIO(content_data),
            length=len(content_data))
        record.content_length = record.content_block.length
        warc.records.append(record)

        bytes(warc)


class TestReading(unittest.TestCase):
    test_dir = os.path.join('example')

    def test_read_c1_warcinfo(self):
        warc = model.WARC()

        warc.load(os.path.join(self.test_dir, 'c1_warcinfo.warc'))
        self.assertEqual(1, len(warc.records))

        record = warc.records[0]

        self.assertEqual('0.18', record.header.version)
        self.assertEqual('warcinfo', record.warc_type)
        self.assertEqual(400, record.content_length,
            'newline between named fields and 0-length payload not accounted'
            ' for')
        self.assertEqual(5, len(record.header.fields))

        content_block = record.content_block
        self.assertEqual('Heritrix 1.12.0 http://crawler.archive.org',
            content_block.fields['software'])
        self.assertEqual(
            'http://www.archive.org/documents/WarcFileFormat-0.18.html',
            content_block.fields['conformsTo'])

    def test_read_at_warc(self):
        warc = model.WARC()

        warc.load(os.path.join(self.test_dir, 'at.warc'))
        self.assertEqual(8, len(warc.records))

        record = warc.records[0]
        file_obj = record.content_block.binary_block.get_file()
        self.assertTrue(file_obj)

    def test_read_at_warc_gzip(self):
        warc = model.WARC()

        warc.load(os.path.join(self.test_dir, 'at.warc.gz'))
        self.assertEqual(8, len(warc.records))

        record = warc.records[0]
        file_obj = record.content_block.binary_block.get_file()
        self.assertTrue(file_obj)

    def test_read_at_warc_memory_gzip(self):
        warc = model.WARC()

        data = io.BytesIO()

        with open(os.path.join(self.test_dir, 'at.warc.gz'), 'rb') as file:
            data.write(file.read())

        data.seek(0)

        file_obj = gzip.GzipFile(fileobj=data)
        warc.read_file_object(file_obj)
        self.assertEqual(8, len(warc.records))

        record = warc.records[0]
        file_obj = record.content_block.binary_block.get_file()
        self.assertTrue(file_obj)

    def test_warc_to_bytes(self):
        warc = model.WARC()

        warc.load(os.path.join(self.test_dir, 'at.warc'))
        bytes(warc)



'''Version info'''

short_version = '2.2'
'''Short version in the form of N.N'''

__version__ = short_version + '.5'
'''Version in the form of N.N[.N]+[{a|b|c|rc}N[.N]+][.postN][.devN]'''


__all__ = ['short_version', '__version__']


if __name__ == '__main__':
    print(__version__, end='')



from distutils.core import setup

import warcat.version

setup(name='Warcat',
    version=warcat.version.__version__,
    description='Tool and library for handling Web ARChive (WARC) files.',
    author='Christopher Foo',
    author_email='chris.foo@gmail.com',
    url='https://github.com/chfoo/warcat',
    packages=[
        'warcat',
        'warcat.model',
    ],
    install_requires=[
        'isodate',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Topic :: System :: Archiving',
    ],
)
